/************************************************************************
 Author: Umit Ogras  (uogras@ece.cmu.edu)
 File name: free_packet_delay.cpp
 Date Written:  <Mon Feb 14 19:16:26 2005>
 Last Modified:
 Description: Read traffic and architecture config files and find the
              free packet delay.
************************************************************************/

#include <iostream>
#include <fstream>
#include <cstdio>
#include <vector>

#define MAX_LINE 1024
using namespace std;
unsigned int reverse_dir(unsigned int curr_dir);

// Parse the config file for the network.
double free_packet_delay(int row, int col, char *traf_config, char *arch_config) {
  
  bool verbose = 0;                             // The switch that controls amount of displayed information
  
  ifstream r_file(arch_config);                 // Open the architecture file
  if (!r_file.is_open() || !r_file.good())
    { cerr <<"The architecture file cannot be opened \n";
      return -1;
    }  
  
  // Initialize the arrival rate matrices
  double arr_rate [row] [col] [6] [6];    
  double arr_rate_agg [row] [col] [6];   
  double arr_rate_agg1 [row] [col];
  for (int i=0;i<row;i++)
    for (int j=0;j<col;j++){
      arr_rate_agg1[i][j]=0;
      for (int k=0;k<6;k++) {
        arr_rate_agg[i][j][k]=0;
        for (int l=0;l<6;l++)
          arr_rate[i][j][k][l]=0;
      }
    }

  // Data structures for routing table
  vector<vector<unsigned int> > routing_table;
  vector<unsigned int> routing_row;
 
  // Initialize the long_range links
  int * RLink; 
  RLink = new int [row*col];
  for (unsigned int i=0;i<row*col;i++)
    RLink[i]=-1;                                   // Initially there are no long_range links
  
  // Initialization for the loop where the architecture file is parsed 
  int id = -1;                                     // Initialize current and destination IDs
  int dst_id=-1;
  char input_line[MAX_LINE];                       // Variable used to parse the input files
  char * sub_string;

  // Parse the architecture file and record the long_range links and the routing table
  while (r_file.getline(input_line, MAX_LINE, '\n')) {
    // line starting with "#" are comments   
    if (input_line[0] == '#')	
      continue;
    
    if (strstr(input_line, "@NODE")) {
      sub_string = strstr(input_line, "@NODE");
      sub_string += strlen("@NODE");
      
      if (!sscanf(sub_string, "%d", &id))
        return -1;
      
      continue;
    } // if @NODE
    
    if ( strstr(input_line, "random_link") ) {
      sub_string = strstr(input_line, "random_link");
      sub_string += strlen("random_link");
      if ( !sscanf(sub_string,"%d",&dst_id) )
        return -1;
      // Found a long_range link (named as random_link for historical reasons) 
      RLink[id]=dst_id;
      RLink[dst_id]=id;

      continue;
    } // if random_link
    
    if (strstr(input_line, "routing_scheme")) {             // Parse the routing table
      sub_string = strstr(input_line, "routing_scheme");
      sub_string += strlen("routing_scheme");
      sub_string = strstr(sub_string,"table");
      sub_string += strlen("table ");
      for (unsigned int i=0; i<strlen(sub_string); i+=2){
        unsigned int new_dir;
        sscanf(&sub_string[i],"%d",&new_dir);
        routing_row.push_back(new_dir);
      } // for 

      routing_table.push_back(routing_row);
      routing_row.erase(routing_row.begin(),routing_row.end());
      
      continue;
    } // if routing_scheme

  } // while
 
  // The architecture file is parsed. The long_range links and routing table are constructed
  r_file.close(); 
 
  // Next parse the traffic file to extract the rate information for each link
  // Open the traffic file
  ifstream in_file(traf_config);
  if (!in_file.is_open() || !in_file.good())
    { cerr <<"Traffic file cannot be opened \n";
      return -1;
    }
  
  double free_delay=0;                                    // Initialize the free packet delay
  while (in_file.getline(input_line, MAX_LINE, '\n')) {
    // line starting with "#" are comments  
    if (input_line[0] == '#')
      continue;
    
    if (strstr(input_line, "@NODE")) {
      sub_string = strstr(input_line, "@NODE");
      sub_string += strlen("@NODE");
      
      if (!sscanf(sub_string, "%d", &id))
        return -1;
      continue;
    }
    
    if (strstr(input_line, "packet_to_destination_rate")) {     
      sub_string = strstr(input_line, "packet_to_destination_rate");     
      sub_string +=strlen("packet_to_destination_rate\t");    
      double rate = 0;
      if (!sscanf(sub_string, "%d\t%lf", &dst_id, &rate)) 
        return -1;     
      if (rate > 1.) {
        cerr << "Rate is larger than 1! \n";
        return -1;
      }      

      unsigned int src_x,src_y,dst_x,dst_y;
      unsigned int curr_x,curr_y,curr_dir;
      int nid=-1;
      src_y = id % row;       
      src_x = (int) (id/row);     
      dst_y = dst_id % row;     
      dst_x = (int) (dst_id/row);      
  
      unsigned int nxt_dir=routing_table[src_y+src_x*row][dst_id];         
      arr_rate[src_x][src_y][0][nxt_dir] += rate;
      
      curr_x = src_x;
      curr_y = src_y;
      unsigned int hops=0;
      while ( (curr_x != dst_x) || (curr_y != dst_y) ) {	     
        // Continue until the curr and dst positions are the same
        curr_dir = routing_table[curr_y+curr_x*row][dst_id];        
        switch (curr_dir) {
        case 1:
          curr_y += 1;
          break;
        case 2:
          curr_y -= 1;
          break; 
        case 3:
          curr_x += 1;         
          break;
        case 4:
          curr_x -= 1;
          break;
        case 5:         
          nid = RLink[curr_y+curr_x*row];
          curr_y = nid % row;
          curr_x = (int) (nid/row);
          break;
        } // Switch
       
        unsigned int rev_dir = reverse_dir(curr_dir);       
        nxt_dir = routing_table[curr_y+curr_x*row][dst_id];      
        arr_rate[curr_y][curr_x][rev_dir][nxt_dir] += rate;
        hops++;
       
       }  // while	    
      
      free_delay += rate*hops;
      continue;
    }
    
    // un-recognized options
    if (strlen(input_line) > 0) {
      cerr <<"Unrecognized option: " << input_line << endl;
      continue;
    }

  } //   while (in_file.getline(input_line, MAX_LINE, '\n'))
  
  // The traffic file is parsed. The traffic rate matrices are constructed
  in_file.close();

  for (int i=0;i<row;i++)
    for (int j=0;j<col;j++) {
      for (int k=0;k<6;k++) {
        for (int l=0;l<6;l++) {         
          arr_rate_agg[i][j][l] += arr_rate[i][j][k][l];
        }
      }      
    } // for (int j=0;j<col;j++)
 
  for (int i=0;i<row;i++)
    for (int j=0;j<col;j++) {
      for (int k=0;k<6;k++) {
        if (verbose)
          cout << arr_rate_agg[i][j][k] <<"\t";
        arr_rate_agg1[i][j] += arr_rate_agg[i][j][k];
        // Check here the conditions for criticality
        if ( arr_rate_agg1[i][j] >= 0.5 ) {
          printf("\nWARNING: Criticality reached at node (%d,%d) \n",i,j);
          printf("\t Try scaling down the packet generation rates\n");
        }
      }
      if (verbose)
        cout <<endl;
    }  // for (int j=0;i<col;j++)
  
  if (verbose)
    cout << "The free packet delay:" << free_delay <<endl;

  return free_delay;
}

unsigned int reverse_dir(unsigned int curr_dir) {
  unsigned int rdir;
  switch (curr_dir) {
  case 0:
    rdir = 0;
    break;
  case 1:
    rdir = 2;
    break;
  case 2:
    rdir = 1;
    break;
  case 3:
    rdir = 4;
    break;
  case 4:
    rdir = 3;
    break;
  case 5:
    rdir = 5;
    break;
  }
  return rdir;
}
